def dice_score(input, target):
    '''
    input and target: Tensors
    '''
    smooth = 1.
    iflat = input.view(-1)
    tflat = target.view(-1)
    intersection = (iflat * tflat).sum()
    return (2.*intersection+smooth) / (iflat.sum() + tflat.sum() + smooth)



# Train skull stripper UNET using new data
import os
import torch; torch.cuda.empty_cache()
from torch.utils.data import DataLoader
import torch.optim as optim
from torch.optim.lr_scheduler import StepLR
from torchvision import transforms
import numpy as np
from dataset import SkullStripperDataset, load_data
from loss import DiceLoss
from metrics import dice_score
import argparse

def train_skullstripper(data_path,
                        num_train_sample,
                        batch_size=256,
                        lr=0.01,
                        lr_step=100,
                        modality=None,
                        num_epochs=100,
                        skpath=None,
                        save_path='weights'
                        ):
    
    print(f'Starting training: lr: {lr}, batch_size: {batch_size}, modality: {modality}, number of mice samples: {num_train_sample}')

    # where to save the model
    modelpath = os.path.join(save_path, modality+'.pth')

    # Load the data
    src_train, msk_train, src_val, msk_val, _, _ = load_data(data_path, num_train_sample, modality=modality)

    # Transform both images and masks ...
    trans = transforms.Compose([transforms.Resize((225,225)),transforms.CenterCrop(256), transforms.ToTensor()])

    train = SkullStripperDataset(src_train, msk_train, transform=trans, augmentation=True)
    val = SkullStripperDataset(src_val, msk_val, transform=trans, augmentation=False)

    training = DataLoader(train, batch_size=batch_size, shuffle=False)
    validating = DataLoader(val, batch_size=batch_size, shuffle=False)

    # Load UNET model and weights from skull-stripper paper
    model = torch.hub.load('mateuszbuda/brain-segmentation-pytorch', 'unet',
                        in_channels=3, out_channels=1, init_features=32,
                        pretrained=False)
    model.load_state_dict(torch.load(skpath, map_location=torch.device('cpu')))
    if torch.cuda.is_available():
        model.cuda()

    loss_f = DiceLoss()
    optimizer = optim.Adam(model.parameters(), lr=lr)
    # scheduler = StepLR(optimizer, step_size=lr_step, gamma=0.1)

    # Metric placeholders
    train_loss, val_loss, train_dice_scores, val_dice_scores = [], [], [], []
    best_dice_score = 0

    # Train the model
    for epoch in range(num_epochs):

        # Train
        model.train()
        l_temp, train_dice_score_temp = [], []
        for inputs, labels in training:
            optimizer.zero_grad()
            if torch.cuda.is_available():
                inputs = inputs.cuda()
                labels = labels.cuda()
            predict = model(inputs)

            loss = loss_f(predict, labels)
            
            loss.backward()
            optimizer.step()

            # Saving metrics
            l_temp.append(loss.item())
            train_dice_score_temp.append(dice_score(labels, predict).item())
        
        train_loss.append(l_temp)
        train_dice_scores.append(train_dice_score_temp)

        # scheduler.step()

        # Validation
        model.eval()
        l_temp, val_dice_score_temp = [], []
        for inputs, labels in validating:
            with torch.no_grad():
                if torch.cuda.is_available():
                    inputs = inputs.cuda()
                    labels = labels.cuda()
                predict = model(inputs)
                loss = loss_f(predict, labels)

                # Saving metrics
                l_temp.append(loss.item())
                val_dice_score_temp.append(dice_score(labels, predict).item())
            
        val_loss.append(l_temp)
        val_dice_scores.append(val_dice_score_temp)
            
        print('epoch [{}/{}], loss train:{:.4f}, val:{:.4f} || dice score train: {:.4f}, val: {:.4f}'.format(epoch+1, num_epochs, \
                    np.mean(train_loss[-1]), np.mean(val_loss[-1]), \
                    np.mean(train_dice_scores[-1]), np.mean(val_dice_scores[-1])), \
                )
        # Saving trained model
        if best_dice_score < np.mean(val_dice_scores[-1]):
            best_dice_score = np.mean(val_dice_scores[-1])
            torch.save(model.state_dict(), modelpath)

    return train_loss, val_loss, train_dice_scores, val_dice_scores

if __name__ == "__main__":
    data_path = '/projects/compsci/USERS/frohoz/msUNET/train/dataset/'
    parser=argparse.ArgumentParser()
    parser.add_argument("-m", "--modality", type=str, required=False)
    parser.add_argument("-n", "--num_train_sample", type=int, required=True)
    args=parser.parse_args()
    print(f'Starting training for modality {args.modality}')
    train_skullstripper(data_path,
                        num_train_sample=args.num_train_sample,
                        batch_size=64,
                        lr=0.01,
                        lr_step=100,
                        modality=args.modality,
                        num_epochs=100,
                        # skpath='/home/ghanba/ssNET/SkullStripperUNET/weights/'+args.modality+'.pth',
                        skpath='/home/ghanba/ssNET/skull-stripper/paper_weights/skull-stripper-paper.pth'
                        )



# Copyright The Jackson Laboratory, 2021
# authors: Jim Peterson, Abed Ghanbari

import sys
import os
import argparse
from argparse import RawTextHelpFormatter
from datetime import datetime

from version import version

_desc = "mmar - (MRI motion artifact removal) Identify and remove frames showing motion artifacts"

_epilog = ""
_epilog += '-------------------------------------------------------------------------------------------------------' + os.linesep
_epilog += 'The <input_scan> is an image file (tif or nii.gz) containing a list of multi-frame slice images' + os.linesep
_epilog += 'The "bvecs" and "bvals" files contain the corresponding diffusion metadata.  Both are tab-separated files.' + os.linesep
_epilog += 'The bvals file is a single record, while the bvecs file is three records for the x, y, z components separated.' + os.linesep
_epilog += 'Intensity values in the bvals file that are less than "bvals_threshold" are assumed to be non-diffusion' + os.linesep
_epilog += 'frames, indicating the start of a new scan.' + os.linesep
_epilog +=  os.linesep
_epilog += 'When run successfully, the output directory will be populated with the following:' + os.linesep
_epilog += '    <output_dir>' + os.linesep
_epilog += '        <log.txt>' + os.linesep
_epilog += '        excluded_frames.csv' + os.linesep
_epilog += '        <scan_directory>' + os.linesep
_epilog += '            <slice_directory>' + os.linesep
_epilog += '                image.nii.gz' + os.linesep
_epilog += '                image.tif' + os.linesep
_epilog += '                mask.tif' + os.linesep
_epilog += '                meta.bvals' + os.linesep
_epilog += '                meta.bvecs' + os.linesep
_epilog += '                profile.jpg' + os.linesep
_epilog += ' ' + os.linesep
_epilog += 'The file "excluded_frames.csv" contains comma-separated values, with one line per slice, listing' + os.linesep
_epilog += 'the frames that were excluded from the original image.  In each record, the first two columns list the ' + os.linesep
_epilog += 'scan index and the slice index, followed by rejected frames.  NOTE: All indices are 0-based.' + os.linesep
_epilog += ' ' + os.linesep
_epilog += 'If the input image contains only one scan the the <scan> directory will be dropped, otherwise' + os.linesep
_epilog += 'the first and second scan directories will be named "small_shell", and "large_shell", and' + os.linesep
_epilog += 'subsequent directories, "scan_03", "scan_04"..., etc.' + os.linesep
_epilog += os.linesep
_epilog += 'Each slice directory contains:' + os.linesep
_epilog += '    image.nii.gz - multi-frame slice image with rejected frames removed' + os.linesep
_epilog += '    image.tif - same as above, in tif format' + os.linesep
_epilog += '    meta.bvals - copy of original bvals file with rejected frames removed' + os.linesep
_epilog += '    meta.bvecs - copy of original bvecs file with rejected frames removed' + os.linesep
_epilog += '    mask.tif - mask used to separate tissue area from background' + os.linesep
_epilog += '    profile.jpg - profile of average intensities in foreground and background across all original frames' + os.linesep
_epilog += '-------------------------------------------------------------------------------------------------------' + os.linesep
_epilog += os.linesep



class command_line_params:
    scan_file = ""
    input_dir = "./"
    output_dir = "./results/"
    bvals_file = ""
    bvecs_file = ""
    scan_path = ""
    bvals_path = ""
    bval_thresh = 10.0
    bvecs_path = ""
    rejected_frames_file = ""
    fg_thresh = 6.0
    bg_thresh = 3.5
    log_file = "log.txt"
    log_file_path = ""
    verbose = False



    def __init__(self):
        parser = argparse.ArgumentParser(description=_desc, formatter_class=RawTextHelpFormatter, epilog=_epilog)
        parser.add_argument("input_scan", help="multi-frame 3D scan")
        parser.add_argument('-i', metavar='input_dir', default=self.input_dir, help="directory containing input files. Default="+self.input_dir)
        parser.add_argument('-o', metavar='output_dir', default=self.output_dir, help="directory for results. Default="+self.output_dir)
        parser.add_argument('-bvecs', metavar='bvecs_file', default=self.bvecs_file, help="input bvecs file - When not specified the scan file base + '.bvecs' is used")
        parser.add_argument('-bvals', metavar='bvals_file', default=self.bvals_file, help="input bvals file - When not specified the scan file base + '.bvals' is used")
        parser.add_argument('-n', metavar='bvals_threshold', type=float, default=self.bval_thresh, help="non-diffusion bval threshold, to determine begining of scan (e.g. 10.0)")
        parser.add_argument('-r', metavar='rejected_frames', default=self.rejected_frames_file, help="When specified, overriding automatic detection")
        parser.add_argument('-fg', metavar='fg_thresh', type=float, default=self.fg_thresh, help="foreground threshold for MAD algorithm (e.g. 6.0)")
        parser.add_argument('-bg', metavar='bg_thresh', type=float, default=self.bg_thresh, help="background threshold for MAD algorithm (e.g. 3.5)")
        parser.add_argument('-log', metavar='log_file', type=str, default=self.log_file, help="log file to be created in the output directory")
        parser.add_argument('-v', dest='verbose', action='store_true', help="print messages to console")

        args = parser.parse_args()

        self.scan_file = args.input_scan
        self.input_dir = args.i
        self.output_dir = args.o
        self.bvecs_file = args.bvecs
        self.bvals_file = args.bvals
        self.bval_thresh = args.n
        self.rejected_frames_file = args.r
        self.fg_thresh = args.bg
        self.fg_thresh = args.fg
        self.log_file = args.log
        self.verbose = args.verbose
        return

    def check_parameters(self):
        i = self.scan_file.rfind('.nii')
        if i<= 0:
            i = self.scan_file.rfind('.tif')
        if i <= 0:
            print(f'error: input DTI scan must be either a "tif" or "nii" file: {self.scan_file}')
            sys.exit(-1)

        base_input_file = self.scan_file[:i]
        if base_input_file.endswith("_0"):
            base_input_file = base_input_file[:-2]
        if not self.bvecs_file:
            self.bvecs_file = base_input_file + ".bvecs"
        if not self.bvals_file:
            self.bvals_file = base_input_file + ".bvals"

        self.scan_path = os.path.join(self.input_dir, self.scan_file)
        self.bvals_path = os.path.join(self.input_dir, self.bvals_file)
        self.bvecs_path = os.path.join(self.input_dir, self.bvecs_file)
        self.log_file_path = os.path.join(self.output_dir, self.log_file)

        if not os.path.isfile(self.scan_path):
            print(f"\nerror: input image file does not exist ({self.scan_path})\n")
            sys.exit(-1)
        if not os.path.isfile(self.bvals_path):
            print(f"\nerror: input bvals file does not exist ({self.bvals_path})\n")
            sys.exit(-1)
        if not os.path.isfile(self.bvecs_path):
            print(f"\nerror: input bvecs file does not exist ({self.bvecs_path})\n")
            sys.exit(-1)

        return


    def params_message(self):
        cumulative_string = os.linesep
        cumulative_string += f"mmar version {version().str}{os.linesep}"
        cumulative_string += f"    date run: {datetime.today().strftime('%m/%d/%Y %H:%M')}{os.linesep}"
        cumulative_string += f"    scan = {self.scan_file}{os.linesep}"
        cumulative_string += f"    input_dir = {self.input_dir}{os.linesep}"
        cumulative_string += f"    output_dir = {self.output_dir}{os.linesep}"
        cumulative_string += f"    bvals_file = {self.bvals_file}{os.linesep}"
        cumulative_string += f"    bvecs_file = {self.bvecs_file}{os.linesep}"
        cumulative_string += f"    bvals_threshold = {self.bval_thresh}{os.linesep}"
        cumulative_string += f"    fg_thresh = {self.fg_thresh}{os.linesep}"
        cumulative_string += f"    bg_thresh = {self.bg_thresh}{os.linesep}"
        if self.rejected_frames_file:
            cumulative_string += f"    rejected_frames = {self.rejected_frames_file}{os.linesep}"
        return cumulative_string




import os
import sys
import numpy as np
import pandas as pd
import tifffile as tiff
from skimage.restoration import estimate_sigma
from collections import Counter
import matplotlib.pyplot as plt

from ..mmar import process_slice



def feature_extractor(image, mask):
    # individual image feature extractor
    features = [
        np.mean(image), 
        np.std(image), 

        np.mean(image[mask]), 
        np.std(image[mask]), 
        np.mean(image[mask]) / np.std(image[mask]),
        
        np.mean(image[mask==False]), 
        np.std(image[mask==False]), 
        np.mean(image[mask==False]) / np.std(image[mask==False]),
        
        estimate_sigma(image),
        
        psnr(image, image*mask),
        psnr(image, image*(1-mask))
        ]
    return features

def extract_features_for_all(parent_folder, mouse_list):

    features = {}
    for mouse in mouse_list:
        print(mouse)
        tif_file = os.path.join(parent_folder, mouse, 'dti_large_0.tif')
        if not os.path.isfile(tif_file):
            features[mouse]=[]
            continue

        img = tiff.imread(tif_file)[1:,:,:,:]
        nframes, nslices, nx, ny = img.shape
        nf_static = 11 # number of static features
        nf_dynamic = 7 # number of dynamic features -- has some information beyond 2D image
        features[mouse] = np.empty((nframes, nslices, nf_static+nf_dynamic))
        for i in range(nframes):
            for j in range(nslices):
                mask = process_slice.create_mask(img[i,j,:,:], fg=False)[0]
                features[mouse][i,j,:nf_static] = feature_extractor(img[i,j,:,:], mask)
            # add general slice feature to each individual frame // based on Median Absloute Deviation

        for j in range(nslices):
            allframes_data_median = np.median(features[mouse][:,j,0])
            allframes_fg_data_median = np.median(features[mouse][:,j,2])
            allframes_bg_data_median = np.median(features[mouse][:,j,4])
            allframes_MAD = np.median([np.abs(features[mouse][:,j,0] - allframes_data_median)])
            allframes_fg_MAD = np.median([np.abs(features[mouse][:,j,2] - allframes_fg_data_median)])
            allframes_bg_MAD = np.median([np.abs(features[mouse][:,j,4] - allframes_bg_data_median)])
            for i in range(nframes):
                features[mouse][i,j,nf_static] = allframes_fg_data_median
                features[mouse][i,j,nf_static+1] = allframes_fg_MAD
                features[mouse][i,j,nf_static+2] = allframes_bg_data_median
                features[mouse][i,j,nf_static+3] = allframes_bg_MAD
                features[mouse][i,j,nf_static+4] = allframes_MAD
                features[mouse][i,j,nf_static+5] = allframes_data_median
                features[mouse][i,j,nf_static+6] = j    

            
    return features

def psnr(img1, img2):
    '''
    Peak Signal to Noise Ratio
    '''
    mse = np.mean((img1 - img2) ** 2)
    return 20*np.log10(np.max(img1) / np.sqrt(mse))

def find_excel_file(folder_path):
    file_list = os.listdir(folder_path)
    excel_file_list = []
    for file in file_list:
        if file.endswith(".xlsx") and not file.startswith("~"):
            excel_file_list.append(file)
    return excel_file_list

def get_rejected_frames(parent_folder, mouse_list):

    rejected = {}
    for mouse in mouse_list:
        rejected_frames_file = find_excel_file(os.path.join(parent_folder,mouse))
        assert len(rejected_frames_file)==1, f"more than one excel file in {mouse}"
        df = pd.read_excel(os.path.join(parent_folder,mouse,rejected_frames_file[0]))
        rejected[mouse] = [[]]*17
        for i, val in enumerate(df[df.columns[0]]):
            if isinstance(val, int): 
                if isinstance(df.iloc[i,1], str):
                    rejected[mouse][val-1]=[int(i)-2 for i in df.iloc[i,1].split(',') if len(i)>0]
                elif isinstance(df.iloc[i,1], int):
                    rejected[mouse][val-1]=df.iloc[i,1]-2
    return rejected


def set_to_one(A, B):
    for i in range(len(A)):
        if A[i] != []:
            if isinstance(A[i], int):
                B[A[i]][i] = 1
                continue
            for j in A[i]:
                B[j][i] = 1
    return B

def plot_features(X, Y, n):
    fig = plt.figure(dpi=75, figsize=(8,8))

    k = 1

    for j in range(X.shape[1]):
        plt.subplot(5,5,k)
        counter = Counter(Y)
        for label, _ in counter.items():
            row_ix = np.where(Y == label)[0]
            plt.scatter(X[row_ix, n], X[row_ix, j], label=str(label))
        k+=1

    plt.legend()
    plt.show()


from skimage.color import rgb2gray
from skimage.filters import threshold_triangle, threshold_otsu
from skimage.morphology import binary_closing
from scipy import ndimage
import SimpleITK as sitk
import numpy as np

def bg_estimation(f_path, threshold_algo='triangle_otsu'):
    '''
    Estimates the background and foreground intensities of a stack of images.
    The background is estimated using a thresholding algorithm, and the foreground is whatever that remains.
    
    Parameters
    ----------
    f_path : str
        Path to the image stack file.
    threshold_algo : str, optional
        Thresholding algorithm to use. Options are ['otsu', 'triangle', 'triangle_otsu'].
        The default is 'triangle_otsu'.
    
    Returns
    -------
    normalized_fg : ndarray (frames, slices)
        Average intensities of the forground 

    background_fg : ndarray (frames, slices)
        Average intensities of the background 

    '''
    # Read image from path file
    img_stack = sitk.GetArrayFromImage(sitk.ReadImage(f_path))

    # threshold to separate foreground and background
    frames, slices, height, width = img_stack.shape
    bg_mask = np.zeros_like(img_stack)
    threshold = np.zeros_like(img_stack)
    for ifr in range(frames):
        for iz in range(slices):
            bg_mask[ifr, iz,:,:], threshold[ifr, iz,:,:] = estimate_background(img_stack[ifr, iz,:,:], threshold_algo=threshold_algo)


    normalized_fg = np.zeros((frames, slices))
    normalized_bg = np.zeros((frames, slices))
    for ifr in range(frames):
        for iz in range(slices):
            normalized_bg[ifr, iz] = np.sum(img_stack[ifr, iz,:,:] * (1-bg_mask[ifr, iz,:,:])) / np.sum((1-bg_mask[ifr, iz,:,:]))
            normalized_fg[ifr, iz] = np.sum(img_stack[ifr, iz,:,:] * (bg_mask[ifr, iz,:,:])) / np.sum((bg_mask[ifr, iz,:,:]))
    return normalized_fg, normalized_bg

def estimate_background(image, threshold_algo='triangle_otsu',
                        threshold=None):
    """
    Estimates the background and pixel intesnity background theshold for an
    image. Note intensity is calculated using skimage.color.rgb2grey
    Parameters
    ----------
    image: array-like, (height, width, n_channels)
        Image whose background to estimate.
    threshold_algo: str, ['otsu', 'triangle', 'triangle_otsu']
        Thresholding algorithm to estimate the background.
        'otsu': skimage.filters.threshold_otsu
        'triangle': skimage.filters.threshold_triangle
        'triangle_otsu': .9 * triangle + .1 * otsu
    threshold: None, float, int
        User provided threshold. If None, will be estimated using one
        of the thresholding algorithms.
    Output
    ------
    background_mask, threshold
    background_mask: array-like, (height, width)
        The True/False mask of estimated background pixels.
    threshold: float
        The (lower bound) backgound threshold intensity for the image.
    """

    grayscale_image = rgb2gray(image)

    if threshold is None:
        if threshold_algo == 'otsu':
            threshold = threshold_otsu(grayscale_image)

        elif threshold_algo == 'triangle':
            threshold = threshold_triangle(grayscale_image)

        elif threshold_algo == 'triangle_otsu':
            triangle = threshold_triangle(grayscale_image)
            otsu = threshold_otsu(grayscale_image)

            threshold = .9 * triangle + .1 * otsu

        else:
            raise ValueError('threshold_algo = {} is invalid argument'.\
                  format(threshold_algo))

    background_mask = grayscale_image > threshold

    return close_and_fill(background_mask), threshold

def close_and_fill(image):
    '''
    This function closes and fills a binary image.
    
    Parameters
    ----------
    image : numpy.ndarray
        A binary image.
        
    Returns
    -------
    numpy.ndarray
        A binary image.
    '''
    return ndimage.binary_fill_holes(binary_closing(image))



# Copyright The Jackson Laboratory, 2021
# authors: Jim Peterson, Abed Ghanbari

import os
from PIL import Image
import numpy as np
import SimpleITK as sitk
import tifffile as tiff # we might be able to save tif files using only sitk ...
import matplotlib.pyplot as plt

def write_all_files(output_dir, scan_list, log):
    '''
    Write all output files.  this includes image, mask, bval, and bvec files for all slices, for all scans.

    Parameters:
        output_dir : str
            The path to the base directory.
        scan_list : list
            List of scans.  Each scan is a list of slices, each of which is of type "class multi_frame_slice"
    Return:
        None
    '''
    log.add_msg(f"{os.linesep}modified scans:")

    if not os.path.isdir(output_dir):
        try:
            os.mkdir(output_dir)
        except OSError as error:
            log.add_msg(f"Could not create output directory: {output_dir} ({error})", is_error_msg=True)

    scan_index = 0
    for scan in scan_list:
        if len(scan_list) == 1:
            scan_dir = ""
        else:
            if scan_index == 0:
                scan_dir = "small_shell"
            elif scan_index == 1:
                scan_dir = "large_shell"
            else:
                scan_dir = f"scan_{scan_index}"
            scan_index += 1
        scan_path = os.path.join(output_dir, scan_dir)
        log.add_msg(" ")
        write_scan_files(scan, scan_path, log)

    excluded_frames_file = os.path.join(output_dir, "excluded_frames.csv")
    write_rejected_frames_file(excluded_frames_file, scan_list)

    return


def write_scan_files(scan, scan_path, log):
    if not os.path.isdir(scan_path):
        try:
            os.mkdir(scan_path)
        except OSError as error:
            log.add_msg(f"Could not create output directory: {scan_path} ({error})", is_error_msg=True)

    slice_index = 0
    for slice in scan:
        slice_dir = "Z"+str(slice_index).zfill(3)
        slice_path = os.path.join(scan_path, slice_dir)
        write_slice(slice, slice_path, log)
        slice_index += 1
    return


def write_slice(slice, slice_path, log):
    if not os.path.isdir(slice_path):
        try:
            os.mkdir(slice_path)
        except OSError as error:
            log.add_msg(f"Could not create output directory: {slice_path} ({error})", is_error_msg=True)

    bval_path = os.path.join(slice_path, "meta.bvals")
    bvec_path = os.path.join(slice_path, "meta.bvecs")
    img_path  = os.path.join(slice_path, "image.tif")
    img_path_nii  = os.path.join(slice_path, "image.nii.gz")
    mask_path  = os.path.join(slice_path, "mask.tif")
    profile_path = os.path.join(slice_path, "profile.jpg")
    write_bval_file(bval_path, slice)
    write_bvec_file(bvec_path, slice)
    write_slice_image_as_tiff(img_path, slice)
    log.add_msg(f"    {img_path_nii}")
    write_slice_image_as_nii(img_path_nii, slice)
    write_slice_mask(mask_path, slice)
    write_profile(profile_path, slice)
    return


def write_bval_file(path, slice):
    '''
    Write a bval file as a tab separated list of floats, with one value per frame.

    Parameters:
        path : str
            The path to the file to write.
        slice : class multi_frame_slice
            The slice to write.
    Return:
        None
    '''

    values = slice.bvals
    if len(slice.rejected):
        r = sorted(slice.rejected, reverse=True)
        for i in r:
            del values[i]
    bval_str = ""
    for v in values:
        bval_str += str(v) + '\t'
    bval_str = bval_str[:-1] + '\n'
    with open(path, 'w') as f:
        f.write(bval_str)
    return


def write_bvec_file(path, slice):
    '''
    Write a bvec file as a tab separated list of floats.  For each frame there is a vector, (x, y, z) value.
    The list of vectors is saved with the coordinates separated - a separate list for each coordinate, in the order
    x, y, z.

    Parameters:
        path : str
            The path to the file to write.
        slice : class multi_frame_slice
            The slice to write.
    Return:
        None
    '''
    values = slice.bvecs
    if len(slice.rejected):
        r = sorted(slice.rejected, reverse=True)
        for i in r:
            del values[i]
    x_str = ""
    y_str = ""
    z_str = ""
    for v in values:
        x_str += str(v[0]) + '\t'
        y_str += str(v[1]) + '\t'
        z_str += str(v[2]) + '\t'
    x_str = x_str[:-1] + '\n'
    y_str = y_str[:-1] + '\n'
    z_str = z_str[:-1] + '\n'
    with open(path, 'w') as f:
        f.write(x_str + y_str + z_str)
    return


def write_slice_image_as_tiff(path, slice):
    '''    
    Write a slice to a file.
        
    Parameters
    ----------
    path : str
        The path to the file to write.
    slice : class multi_frame_slice
        The slice to write.
    
    Returns
    -------
    None

    '''
    img = slice.img
    if slice.rejected:
        img = np.delete(img, slice.rejected, 0)
    tiff.imsave(path, img)
    return

def write_slice_image_as_nii(path, slice):
    '''
    Write a slice to a file.

    Parameters
    ----------
    path : str
        The path to the file to write.
    slice : class multi_frame_slice
        The slice to write.

    Returns
    -------
    None

    '''
    img = slice.img
    img_itk = sitk.GetImageFromArray(img)
    sitk.WriteImage(img_itk, path)



def write_slice_mask(path, slice):
    '''
    Write a mask file for a given slice.
    
    Parameters
    ----------
    path : str
        The path to the mask file to be written.
    slice : class multi_frame_slice
        The slice to write the mask for.
    
    Returns
    -------
    None

    '''
    if type(slice.mask) == type(None):
        return
    tiff.imsave(path, slice.mask)
    return




def write_rejected_frames_file(output_file, scan_list):
    '''
    Write the list of rejected frames.

    Parameters:
        output_dir : str
            The path to the directory in which to write the file.
        scan_list : list
            List of scans.  Each scan is a list of slices, each of which is of type "class multi_frame_slice"
    Return:
        None
    '''
    body = "scan, slice, excluded_frames\n"

    for scan_index in range(len(scan_list)):
        scan = scan_list[scan_index]
        for z in range(len(scan)):
            slice = scan[z]
            body += f"{scan_index}, {z}"
            for n in slice.rejected:
                body += f", {n}"
            body += "\n"
    with open(output_file, 'w') as f:
        f.write(body)

    return


def write_profile(file_path, slice):
    '''
    Write the profile of the foreground and background of the specified slice.

    Parameters:
        path : str
            The path to the image file to write.
        slice : class multi_frame_slice
            Profiles of the fg and bg of this slice are generated.
    Return:
        None
    '''

    if type(slice.profile_foreground) == type(None):
        return
    nf = len(slice.profile_foreground)
    x1 = np.linspace(1,nf, num=nf-1)
    fig, ax1 = plt.subplots()
    ax2 = ax1.twinx()
    ax1.plot(x1, slice.profile_foreground[1:], 'g-')
    ax2.plot(x1, slice.profile_background[1:], 'b-')
    ax1.set_xlabel('slice number')
    ax1.set_ylabel('foreground intensity', color='g')
    ax2.set_ylabel('background intensity', color='b')
    plt.savefig(file_path)
    plt.close()





# Copyright The Jackson Laboratory, 2021
# authors: Jim Peterson, Abed Ghanbari

class version:
    major = 1
    minor = 0
    str = None
    def __init__(self):
        self.str = str(self.major) + "." + str(self.minor).zfill(3)



#%%
import os
import numpy as np
import random
from .train import MotionFrameDetector
from .utils import *

def evaluate(n):
    parent_folder = '/Users/ghanba/Documents/ssif_data_jax/'
    mouse_list = [f for f in os.listdir(parent_folder) if not f.startswith('.')]
    mouse_list.sort()

    mouse_list = random.sample(mouse_list[:95], n)

    rejected = get_rejected_frames(parent_folder, mouse_list)

    features = extract_features_for_all(parent_folder, mouse_list)

    X_all = []
    Y_all = []
    for mouse in mouse_list:
        print(mouse)
        X = features[mouse]
        if isinstance(X_all, list):
            X_all = np.expand_dims(X, axis=0)
        else:
            X_all = np.concatenate((X_all, np.expand_dims(X, axis=0)), axis=0)
        
        Y = set_to_one(rejected[mouse], np.zeros((X.shape[0], X.shape[1])))
        if isinstance(Y_all, list):
            Y_all = np.expand_dims(Y, axis=0)
        else:
            Y_all = np.concatenate((Y_all, np.expand_dims(Y, axis=0)), axis=0)
        
    X_all = X_all[:,:,:,:]
    X = X_all.reshape(-1, X_all.shape[-1])
    Y = Y_all.reshape(-1)


    # scatter plot of the dataset
    plot_features(X, Y, 11)

    model = MotionFrameDetector(X, Y, train_balanced=True)
    model.train()
    model.print_results()
    model.plot_confusion_matrices()
    model.plot_roc_curves()
    model.plot_roc_auc_scores()
    model.plot_balanced_scores()
    return model

if __name__ == '__main__':
    evaluate(5)

# %%



#!/usr/bin/env python

# Copyright The Jackson Laboratory, 2021
# authors: Jim Peterson, Abed Ghanbari

'''
This script

To Do:
    - catch possible exception when bval and bvec file entries are converted to numbers
    - catch exceptions when writing output files
    - check for directory separators in log file name

'''
import sys
import os
import numpy as np
import SimpleITK as sitk
import csv

from process_slice import find_bad_frames
from write_output import write_all_files
from command_line_params import command_line_params
from message_log import message_log


'''
import ipywidgets as ipyw
from IPython.display import set_matplotlib_formats
import pandas as pd
'''


class multi_frame_slice:
    img = None
    mask = None
    bvals = None
    bvecs = None
    rejected = None
    profile_background = None
    profile_foreground = None

def main():
    '''
    '''

    #
    # Parse the command line parameters
    #
    p = command_line_params()
    p.check_parameters()
    log = message_log(p.verbose)
    log.add_msg(p.params_message())

    #
    # Open the input files
    #
    bval_list = open_bval_file(p.bvals_path, log)
    bvec_list = open_bvec_file(p.bvecs_path, log)
    img = sitk.GetArrayFromImage(sitk.ReadImage(p.scan_path))

    #
    # Count the number of sequences
    #
    thresh = p.bval_thresh
    frame_starts = []
    for i in range(len(bval_list)):
        if bval_list[i] < thresh:
            frame_starts.append(i)

    frames, slices, height, width = img.shape
    log.add_msg(f"    sequences = {len(frame_starts)}")
    log.add_msg(f"    width = {width}")
    log.add_msg(f"    height = {height}")
    log.add_msg(f"    slices = {slices}")
    log.add_msg(f"    frames = {frames}")
    if not ((frames == len(bval_list)) and (frames == len(bvec_list))):
        log.add_msg("input files do not match in frame count", is_error_msg=True)
        exit(-1)

    #
    # Separate scans
    #
    frame_starts.append(len(bval_list))
    scan_list = []
    for i in range(len(frame_starts)-1):
        bvals = bval_list[frame_starts[i]: frame_starts[i+1]]
        bvecs = bvec_list[frame_starts[i]: frame_starts[i+1]]
        slice_list = []
        for z in range(slices):
            s = multi_frame_slice()
            s.img = img[frame_starts[i]:frame_starts[i+1], z, :, :]
            s.bvals = bvals.copy()
            s.bvecs = bvecs.copy()
            s.rejected = []
            slice_list.append(s)
        scan_list.append(slice_list)
    img = None
    bval_list = None
    bvec_list = None
    frame_starts = None

    #
    # Read the rejected frame file, if specified
    #
    if p.rejected_frames_file:
        if not os.path.isfile(p.rejected_frames_file):
            log.add_msg(f"rejected frames file does not exist: {p.rejected_frames_file}", is_error_msg=True)
            exit(-1)

        lines = []
        with open(p.rejected_frames_file, 'r', encoding='utf-8-sig') as f:
            reader = csv.reader(f)
            lines = list(reader)
        for l in lines:
            if not l[0].isnumeric():
                continue
            n_scan = int(l[0])
            if n_scan < 0 or n_scan >= len(scan_list):
                continue;
            if not l[1].isnumeric():
                continue
            n_slice = int(l[1])
            if n_slice < 0 or n_slice >= len(scan_list[n_scan]):
                continue

            for f in l[2:]:
                if not f.isnumeric():
                    continue
                n_frame = int(f)
                if n_frame < 0 or n_frame >= frames:
                    continue
                scan_list[n_scan][n_slice].rejected.append(n_frame)

    #
    # Loop over all scans and through all slices, processing images
    #
    #     scan = scan_list[n]
    #     slice = scan[z]
    #     image = slice.img
    #     bvals = slice.bvals
    #     bvecs = slice.bvecs
    #
    if not p.rejected_frames_file:
        scan_processed = []
        for i in range(len(scan_list)):
            if i == 0:
                scan_name = "small_shell"
            elif i == 1:
                scan_name = "large_shell"
            else:
                scan_name = f"scan_{scan_index}"
            log.add_msg(f"{os.linesep}scan: {scan_name}")
            slice_processed = []
            for z in range(len(scan_list[i])):
                slice_processed.append(find_bad_frames(scan_list[i][z], p.fg_thresh, p.bg_thresh))
                log.add_msg(f"    slice {z}: {slice_processed[-1].rejected}")
            scan_processed.append(slice_processed)
    
    #
    # Write output files
    #
    write_all_files(p.output_dir, scan_list, log)
    log.save_log(p.log_file_path)

    sys.exit(0)


def open_bval_file(file_path, log):
    '''
    opens a bval file and returns a list of floats.
    
    Parameters
    ----------
    file_path : str
        The path to the file to be opened.
    
    Returns
    -------
    num_list : list
        A list of floats.
    
    Examples
    --------
    >>> open_bval_file("/home/user/bval_file.bvals")
    [0.5, 0.3, 0.2]
    '''
    with open(file_path) as f:
        lines = f.readlines()
    if len(lines) < 1:
        log.add_msg(f"error: input meta file is either empty or can not be opened ({file_path}", is_error_msg=True)
        sys.exit(-1)
    if len(lines) > 1:
        log.add_msg(f"warning: input meta file is not of the proper form ({file_path}", is_error_msg=True)
    line = lines[0]
    if line[-1] == '\n':
        line = line[:-1]
    str_list = line.split('\t')
    num_list = [float(s) for s in str_list]

    return num_list


def open_bvec_file(file_path, log):
    '''
    takes a file path to a bvec file and returns a list of tuples
    containing the x, y, and z components of the bvecs.

    Parameters
    ----------
    file_path : str
        The path to the bvec file.

    Returns
    -------
    vect_list : list of tuples
        A list of tuples containing the x, y, and z components of the bvecs.

    Examples
    --------
    >>> open_bvec_file('/path/to/bvec_file.bvecs')
    [(1.0, 0.0, 0.0), (0.0, 1.0, 0.0), (0.0, 0.0, 1.0)]
    '''
    with open(file_path) as f:
        lines = f.readlines()
    if len(lines) < 3:
        log.add_msg(f"error: input meta file is either empty or not of the proper form ({file_path}", is_error_msg=True)
        sys.exit(-1)
    if len(lines) > 3:
        log.add_msg(f"warning: input meta file is not of the proper form ({file_path}", is_error_msg=True)

    if lines[0][-1] == '\n':
        lines[0] = lines[0][:-1]
    if lines[1][-1] == '\n':
        lines[1] = lines[1][:-1]
    if lines[2][-1] == '\n':
        lines[2] = lines[2][:-1]

    x_str = lines[0].split('\t')
    y_str = lines[1].split('\t')
    z_str = lines[2].split('\t')

    x = [float(s) for s in x_str]
    y = [float(s) for s in y_str]
    z = [float(s) for s in z_str]

    if (len(x) != len(y)) or (len(x) != len(z)) or (len(x) == 0):
        log.add_msg(f"error: input meta file is not of the proper form ({file_path}", is_error_msg=True)
        sys.exit(-1)

    vect_list = [(x[i], y[i], z[i]) for i in range(len(x))]

    return vect_list


if __name__== "__main__":
    main()



# Copyright The Jackson Laboratory, 2021
# authors: Jim Peterson, Abed Ghanbari

import numpy as np

from .create_mask import create_mask

def find_bad_frames(slice, fg_thresh, bg_thresh):
    '''
    Finds the bad frames in a slice.

    Parameters
    ----------
    slice : Slice
        The slice to be processed.

    Returns
    -------
    mask : ndarray
        A boolean array of the same shape as the image.
    slice : Slice
        The modified slice
    '''
    image = slice.img
    bvals = slice.bvals
    bvecs = slice.bvecs
    mask, _ = create_mask(image[0,:,:], fg=False)

    nz = slice.img.shape[0]
    profile_foreground = np.zeros(nz)
    profile_background = np.zeros(nz)
    for iz in range(nz):
        profile_foreground[iz] = np.sum(slice.img[iz,:,:] * (1-mask)) / np.sum((1-mask))
        profile_background[iz] = np.sum(slice.img[iz,:,:] * (mask)) / np.sum((mask))

    slice.mask = mask
    slice.profile_background = profile_background
    slice.profile_foreground = profile_foreground

    e_fg = get_outliers(profile_background, mode='MAD', MAD_thr=fg_thresh)
    e_bg = get_outliers(profile_foreground, mode='MAD', MAD_thr=bg_thresh)
    for n in e_bg:
        if not n in e_fg:
            e_fg.append(n)
    slice.rejected = sorted(e_fg, reverse=False)

    return slice




def get_outliers(data, mode='MAD', r=None, MAD_thr=None):
    '''
    This function identifies outliers in a given dataset.

    Parameters
    ----------
    data : list
        A list of numbers.
    mode : str, optional
        The method used to identify outliers.
        'std' uses standard deviation method.
        'MAD' uses median absolute deviation method.
        The default is 'MAD'.
    r : float, optional
        The ratio used to identify outliers.
        The default is None.
    MAD_thr : float, optional
        The threshold used to identify outliers using median absolute deviation method.
        The default is None, however in [https://www.sciencedirect.com/science/article/abs/pii/S0022103113000668]
        it has been suggested to use 2.5

    Returns
    -------
    outliers : list
        A list of indices of outliers in the given dataset.

    '''
    if mode=='std':
        mean = np.mean(data)
        std = np.std(data)
        outliers = []
        for i in range(1,len(data)):
            if data[i] > mean + r*std or data[i] < mean - r*std:
                outliers.append(i)
    elif mode=='MAD':
        # calculate median and median absolute deviation
        # ref: https://www.sciencedirect.com/science/article/abs/pii/S0022103113000668
        data_median = np.median(data)
        data_mad = np.median([np.abs(data - data_median)])
        # identify outliers
        cut_off = data_mad * MAD_thr
        lower, upper = data_median - cut_off, data_median + cut_off
        outliers = []
        for i in range(1,len(data)):
            if data[i] < lower or data[i] > upper:
                outliers.append(i)
    return outliers




import torch; torch.cuda.empty_cache()
from torchvision import transforms
from PIL import Image
from skimage.transform import resize
from .utils import cvt1to3channels, normalize_image
import os
def output_mask(img):

    skpath = os.path.dirname(__file__)+'/weights/dti.pth'
    model = torch.hub.load('mateuszbuda/brain-segmentation-pytorch', 'unet',
                        in_channels=3, out_channels=1, init_features=32,
                        pretrained=False)
    # load pretrained weights trianed on our data
    model.load_state_dict(torch.load(skpath, map_location=torch.device('cpu')))
    
    for p in model.parameters():
        p.requires_grad = False
    model.eval()

    trans = transforms.Compose([transforms.Resize((225,225)),transforms.CenterCrop(256), transforms.ToTensor()])

    img = normalize_image(img)
    tensor_stack = trans(Image.fromarray(cvt1to3channels(img.astype('uint8')))).unsqueeze(0)
    
    mask_tensor = model(tensor_stack)[0,0,:,:]

    out_trans = transforms.Compose([
                        transforms.CenterCrop((225,225)),
                    ])
    mask_orig = resize(out_trans(mask_tensor).numpy(),(img.shape[0],img.shape[1]))
    return mask_orig




# %%
# from torch.utils.data import Dataset
# import torchvision.transforms.functional as F
import numpy as np
import SimpleITK as sitk
from utils import cvt1to3channels, normalize_image
import random
from PIL import Image
# from torchvision import transforms
import os
from random import seed, shuffle

def load_data(data_path, num_train_sample, modality=None):

    train_path = data_path + 'training/' 
    mask_path = data_path + 'masks/' 
    train_list = sorted(os.listdir(train_path))

    mouse_name = np.unique([f[:24] for f in train_list if f[:6]=='CK_DIF'])
    seed(10)
    shuffle(mouse_name)
    mouse_train = mouse_name[:num_train_sample]
    mouse_val   = mouse_name[num_train_sample:]

    train_list = [i for i in train_list if i[-2:]=='gz']

    src_train, msk_train, fnames_train = [], [], []
    src_val, msk_val, fnames_val = [], [], []

    # Read indiviual "nii.gz" files
    for f in train_list:
        if modality and modality not in f:
                continue
        # Read the entire volume
        training_arr = sitk.GetArrayFromImage(sitk.ReadImage(train_path + f))
        
        # 'noddi' images have 4 dimentions
        if len(training_arr.shape)==4:
            training_arr = training_arr[0,:,:,:]
        
        mask_arr = sitk.GetArrayFromImage(sitk.ReadImage(mask_path + f.replace('data','mask')))

        for image_idx in range(training_arr.shape[0]):
            # Preprocess and transform training data
            input_image_original = training_arr[image_idx, :,:]
            input_image_original = normalize_image(input_image_original)

            # Transform expert mask
            input_mask_original = mask_arr[image_idx, :,:]
            input_mask_original = normalize_image(input_mask_original)
            if f[:24] in mouse_train:
                src_train.append(np.uint8(input_image_original))
                msk_train.append(np.uint8(input_mask_original))
                fnames_train.append((f, image_idx))
            elif f[:24] in mouse_val:
                src_val.append(np.uint8(input_image_original))
                msk_val.append(np.uint8(input_mask_original))
                fnames_val.append((f, image_idx))
            else:
                print(f'{f} is not in train or validation mouse list!')
            
    return src_train, msk_train, src_val, msk_val, fnames_train, fnames_val



# Copyright The Jackson Laboratory, 2021
# authors: Jim Peterson, Abed Ghanbari

import os
import sys
import tifffile as tiff
import pandas as pd
import itertools
import csv
import numpy as np
import matplotlib
from skimage.transform import resize
matplotlib.use("Qt5Agg")
import matplotlib.pyplot as plt
from PyQt5.QtWidgets import QApplication, QComboBox, QWidget, QSlider, QLabel, QPushButton, QListWidget, QGridLayout, QMessageBox, QSizePolicy, QLineEdit, QCheckBox
from PyQt5.QtCore import Qt, QCoreApplication
from PyQt5.QtGui import QIcon
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT as NavigationToolbar
from matplotlib.figure import Figure
import mplcursors

from ..mmar.process_slice import create_mask
from ..ssUNET.generate_output import output_mask

class App(QWidget):
    def __init__(self):
        super().__init__()
        self.title = 'MRI Motion Artifact Removal'
        self.show_unet = True
        # coordinated for the UI
        self.left = 10
        self.top = 10
        self.width = 1000
        self.height = 1000
        self.initUI()

    def initUI(self):
        '''
        initialize the UI
        we use a grid structure for the app
        '''
        self.setWindowTitle(self.title)
        self.setGeometry(self.left, self.top, self.width, self.height)

        # slider widget for the frames
        self.slider = QSlider(Qt.Horizontal, self)
        self.slider.setFocusPolicy(Qt.StrongFocus)
        self.slider.setTickPosition(QSlider.TicksBothSides)
        self.slider.setTickInterval(10)
        self.slider.setSingleStep(1)
        self.slider.setMinimum(0)
        self.slider.setMaximum(100)
        self.slider.setValue(0)
        self.slider.valueChanged[int].connect(self.changeValue_frame) # by each change in slider the changeValue_frame function is called

        self.prntfldr_textbox = QLineEdit(self)
        self.prntfldr_textbox.setText("/Users/ghanba/Documents/ssif_data_jax")
        self.prntfldr_textbox.returnPressed.connect(self.set_parentfldr)

        self.prntfldr_button = QPushButton('Set parent folder')
        self.prntfldr_button.clicked.connect(self.set_parentfldr)


        self.label_frame = QLabel(self) # by each change in slider frame number label will change
        self.label_frame.setText('Frame Number: '+'0')
        self.label_frame.setMinimumWidth(150)

        self.mouse_list_lbl = QLabel(self) # label for list of mouse names
        self.mouse_list_lbl.setText('Mouse Names:')
        self.mouse_list_lbl.setAlignment(Qt.AlignBottom)

        # list of folders found
        self.mouselistbox = QListWidget()
        self.mouselistbox.setMinimumWidth(215)
        self.mouselistbox.setSortingEnabled(True)
        
        self.mouselistbox.itemSelectionChanged.connect(self.updateImageList)

        self.label_z = QLabel(self)
        self.label_z.setText('Slice Number: ')

        # ComboBox
        self.cb = QComboBox()
        self.cb.currentIndexChanged.connect(self.z_selectionchange)

        self.mask_cb = QComboBox()
        self.mask_cb.addItem('UNET')
        self.mask_cb.addItem('Background')
        self.mask_cb.currentIndexChanged.connect(self.set_masks)

        self.image_or_mask = QCheckBox("Mask",self)
        self.image_or_mask.stateChanged.connect(lambda:self.changeValue_frame(self.slider.value()))

        self.label_bg_MAD = QLabel(self)
        self.label_bg_MAD.setText('Background Thr:')
        self.label_bg_MAD.setAlignment(Qt.AlignRight)

        self.bg_MAD = QLineEdit(self)
        self.bg_MAD.setText("3.5")
        self.bg_MAD.setMaximumWidth(30)
        self.bg_MAD.returnPressed.connect(self.updateImageList)

        self.label_fg_MAD = QLabel(self)
        self.label_fg_MAD.setText('Foreground Thr:')

        self.fg_MAD = QLineEdit(self)
        self.fg_MAD.setText("5.5")
        self.fg_MAD.setMaximumWidth(30)
        self.fg_MAD.returnPressed.connect(self.updateImageList)

        # Figure canvas and toolbar
        self.figure = plt.figure()
        self.canvas = FigureCanvas(self.figure)
        self.toolbar = NavigationToolbar(self.canvas, self)
        self.toolbar.setMaximumHeight(30)

        # list of accepted frames
        self.framelist_accepted = QListWidget()
        self.framelist_accepted.setMaximumWidth(60)
        self.framelist_accepted.setSortingEnabled(True)
        self.framelist_accepted.itemDoubleClicked.connect(self.framelist_accepted_double_clicked)
        self.framelist_accepted.itemSelectionChanged.connect(self.framelist_accepted_clicked)

        # list of rejected frames
        self.framelist_rejected = QListWidget()
        self.framelist_rejected.setMaximumWidth(60)
        self.framelist_rejected.setSortingEnabled(True)
        self.framelist_rejected.itemDoubleClicked.connect(self.framelist_rejected_double_clicked)
        self.framelist_rejected.itemSelectionChanged.connect(self.framelist_rejected_clicked)

        # adding labels at the top of the lists
        self.label_list1 = QLabel(self)
        self.label_list1.setText('Accepted')
        self.label_list2 = QLabel(self)
        self.label_list2.setText('Rejected')

        # button for saving the list of rejected frames
        self.save_button = QPushButton('Save list')
        self.save_button.setMaximumWidth(75)
        self.save_button.clicked.connect(self.save_onclick)

        # button for loading rejected frames list 
        self.load_button = QPushButton('Load list')
        self.load_button.setMaximumWidth(75)
        self.load_button.clicked.connect(self.load_onclick)

        self.exit_button = QPushButton('Exit')
        self.exit_button.clicked.connect(self.exitapp)

        ################## Layout ##################
        layout = QGridLayout()
        layout.addWidget(self.prntfldr_textbox,0,1,1,4)
        layout.addWidget(self.prntfldr_button,0,5,1,1)
        layout.addWidget(self.mouse_list_lbl,1,0,1,1)
        layout.addWidget(self.mouselistbox, 2, 0, 22,1)
        layout.addWidget(self.toolbar,21,1,1,5)
        layout.addWidget(self.canvas,2,1,19,5)
        layout.addWidget(self.label_z,22,1,1,1)
        layout.addWidget(self.cb,22,2,1,1)
        layout.addWidget(self.mask_cb,22,5,1,1)
        layout.addWidget(self.image_or_mask,22,4,1,1)
        layout.addWidget(self.label_frame,23,1,1,1)
        layout.addWidget(self.slider,23,2,1,4)
        layout.addWidget(self.label_list1,1, 6,1,1)
        layout.addWidget(self.label_list2,1, 7,1,1)
        layout.addWidget(self.framelist_accepted, 2, 6,21,1)
        layout.addWidget(self.framelist_rejected, 2, 7,21,1)
        layout.addWidget(self.load_button, 23, 6,1,1)
        layout.addWidget(self.save_button, 23, 7,1,1)
        layout.addWidget(self.label_bg_MAD, 24,1,1,1)
        layout.addWidget(self.bg_MAD, 24,2,1,1)
        layout.addWidget(self.label_fg_MAD, 24,3,1,1)
        layout.addWidget(self.fg_MAD, 24,4,1,1)

        self.setLayout(layout)

        self.show()
    def get_filename(self):
        return os.path.join(self.parent_folder, self.mouse_name, 'dti_large_0.tif')

    def set_parentfldr(self):
        if os.path.isdir(self.prntfldr_textbox.text()):
            self.parent_folder = self.prntfldr_textbox.text()
            self.mouse_folders = [f for f in os.listdir(self.parent_folder) if not f.startswith('.')] # lists all folders in the parent folder except the ones starting with "."
            self.mouse_name = self.mouse_folders[0]  # default to first in the list of folders
            self.file_name = self.get_filename()
            self.mouselistbox.clear()
            for i, fldr in enumerate(self.mouse_folders):
                self.mouselistbox.insertItem(i, fldr)
            self.mouselistbox.setCurrentRow(0)
            self.updateImageInit()
            for i in range(self.nz):
                self.cb.addItem(str(i))
            self.cb.setCurrentIndex(0)
            self.slider.setMaximum(self.data.shape[0]-1)
            self.changeValue_frame(0)
        else:
            self.showDialog(msg='Folder does not exist:\n'+self.prntfldr_textbox.text())

    def updateImageInit(self):
        '''
        in first pass of the app loads the first folder's data and creates the mask
        '''
        if os.path.isfile(self.file_name):
            self.data_all = tiff.imread(self.file_name) # reads image data 
            self.nz = self.data_all.shape[1] # sets number of slices
            self.data = self.data_all[:,0,:,:] # sets the current slice data to slice 0
            self.set_masks()
            self.rejected_frames = [set() for i in range(self.nz)] # should be a set of rejected frames data structures
        else:
            msg='File does\'nt exist:\n'+self.file_name
            print('warning:'+msg)
            self.showDialog(msg=msg)

    def updateImageList(self):
        '''
        by selecting mouse updates the app
        '''
        self.mouse_name = self.mouselistbox.currentItem().text() # gets what has been selected in mice list
        self.file_name = self.get_filename()
        self.updateImageInit() # gets the data for the current mouse
        self.changeValue_frame(0) # goes to frame 0
        self.slider.setValue(0) # set the slider value to 0
        self.plot_intensity_profile() # plots the intensity plots for slice 0 and frame 0
        self.update_textboxs() # updates the rejected and accepted frames
        self.cb.setCurrentIndex(0)

    def changeValue_frame(self, value):
        if hasattr(self,'data'):
            self.label_frame.setText('Frame Number: '+str(value))
            
            if self.image_or_mask.isChecked():
                if self.mask_cb.currentText()=='UNET':
                    data_show=self.maskunet
                else:
                    data_show=self.maskdata

            else:
                data_show = self.data[value, :, :]
            
            self.figure.clear()
            ax = self.figure.add_subplot(111)
            
            if value==0 or self.image_or_mask.isChecked():
                ax.imshow(data_show, cmap='gray')
            else:
                # for frames (except the frame 0) we keep the range the same so imshow doesn't adjust contrast automatically
                ax.imshow(data_show, cmap='gray', vmin=np.min(self.data[1:,:,:]), vmax=np.max(self.data[1:,:,:]))
            plt.axis('off')
            self.figure.tight_layout()
            self.canvas.draw()

    def framelist_accepted_double_clicked(self):
        item = self.framelist_accepted.currentItem()
        self.framelist_rejected.addItem(item.text())
        self.framelist_accepted.takeItem(self.framelist_accepted.row(item))
        # if a frame in accepted frames was double clicked add it to set of rejected frames
        self.rejected_frames[self.cb.currentIndex()].add(int(item.text()))

    def framelist_rejected_double_clicked(self):
        item = self.framelist_rejected.currentItem()
        self.framelist_accepted.addItem(item.text())
        self.framelist_rejected.takeItem(self.framelist_rejected.row(item))
        self.rejected_frames[self.cb.currentIndex()].remove(int(item.text()))

    def framelist_accepted_clicked(self, item=None):
        if not item:
            item = self.framelist_accepted.currentItem()
        self.slider.setValue(int(item.text()))

    def framelist_rejected_clicked(self, item=None):
        if not item:
            item = self.framelist_rejected.currentItem()
        if item:
            self.slider.setValue(int(item.text()))

    def z_selectionchange(self):
        self.data = self.data_all[:,self.cb.currentIndex(),:,:]
        self.set_masks()
        self.changeValue_frame(0)
        self.slider.setValue(0)
        self.plot_intensity_profile()
        self.update_textboxs()

    def update_textboxs(self):
        self.framelist_accepted.clear()
        self.framelist_rejected.clear()
        for i in range(self.data.shape[0]):
            if i in self.rejected_frames[self.cb.currentIndex()]:
                self.framelist_rejected.insertItem(i, str(i).zfill(2))
            else:
                self.framelist_accepted.insertItem(i, str(i).zfill(2))

    def save_onclick(self):
        if hasattr(self,'file_name'):
            dict = {}
            for i in range(self.nz):
                dict[str(i)] = list(self.rejected_frames[i])
            with open(os.path.join(os.path.dirname(self.file_name), 'frame_list.csv'), "w") as outfile:
                writer = csv.writer(outfile)
                writer.writerow(dict.keys())
                writer.writerows(itertools.zip_longest(*dict.values()))
    def set_masks(self):
        if hasattr(self,'data'):
            mask_output = output_mask(self.data[0, :, :]) # mask is estimated for the slice 0 and frame 0 since the non directional DTI image is stored in frame 0
            self.maskunet = resize(mask_output, (self.data_all.shape[2], self.data_all.shape[3]), anti_aliasing=True)
            self.maskdata = create_mask(self.data[0,:,:], fg=False)[0] # mask is estimated for the slice 0 and frame 0 since the non directional DTI image is stored in frame 0
            self.changeValue_frame(0)

    def load_onclick(self):
        if hasattr(self,'file_name'):
            csv_file_path = os.path.join(os.path.dirname(self.file_name), 'frame_list.csv')
            if os.path.isfile(csv_file_path):
                df = pd.read_csv(csv_file_path)
                self.rejected_frames = [set() for i in range(self.nz)]
                # recreate rejected_frames dictionary from csv file
                for i in range(self.nz):
                    for frame in list(df.to_dict()[str(i)].values()):
                        if not pd.isna(frame):
                            self.rejected_frames[i].add(int(frame))
                self.update_textboxs()
            else:
                self.showDialog(msg='There is no csv file for this mouse ...')
    
    def calculate_profiles(self):
        '''
        Calculates the profiles of the foreground and background.
        
        Returns
        -------
        profile_foreground : numpy.ndarray
            The profile of the foreground.
        profile_background : numpy.ndarray
            The profile of the background.
        
        Notes
        -----
        The profile of the foreground is calculated by summing the data over the
        pixels in mask, and dividing by the number of pixels in the mask.
        '''
        nz = self.data.shape[0]
        profile_foreground = np.zeros(nz)
        profile_background = np.zeros(nz)
        for iz in range(nz):
            profile_foreground[iz] = np.sum(self.data[iz,:,:] * (self.maskdata)) / np.sum((self.maskdata))
            profile_background[iz] = np.sum(self.data[iz,:,:] * (1-self.maskdata)) / np.sum((1-self.maskdata))
        for i in self.get_outliers(profile_background, mode='MAD', MAD_thr=float(self.bg_MAD.text())):
            self.rejected_frames[self.cb.currentIndex()].add(i)
        for i in self.get_outliers(profile_foreground, mode='MAD', MAD_thr=float(self.fg_MAD.text())):
            self.rejected_frames[self.cb.currentIndex()].add(i)
        return profile_foreground, profile_background

    def get_outliers(self, data, mode='MAD', r=None, MAD_thr=None):
        '''
        This function identifies outliers in a given dataset.
    
        Parameters
        ----------
        data : list
            A list of numbers.
        mode : str, optional
            The method used to identify outliers.
            'std' uses standard deviation method.
            'MAD' uses median absolute deviation method.
            The default is 'MAD'.
        r : float, optional
            The ratio used to identify outliers.
            The default is None.
        MAD_thr : float, optional
            The threshold used to identify outliers using median absolute deviation method.
            The default is None, however in [https://www.sciencedirect.com/science/article/abs/pii/S0022103113000668]
            it has been suggested to use 2.5
        
        Returns
        -------
        outliers : list
            A list of indices of outliers in the given dataset.
        
        '''
        if mode=='std':
            mean = np.mean(data)
            std = np.std(data)
            outliers = []
            for i in range(1,len(data)):
                if data[i] > mean + r*std or data[i] < mean - r*std:
                    outliers.append(i)
        elif mode=='MAD':
            # calculate median and median absolute deviation
            # ref: https://www.sciencedirect.com/science/article/abs/pii/S0022103113000668
            data_median = np.median(data)
            data_mad = np.median([np.abs(data - data_median)])
            # identify outliers
            cut_off = data_mad * MAD_thr
            lower, upper = data_median - cut_off, data_median + cut_off
            outliers = []
            for i in range(1,len(data)):
                if data[i] < lower or data[i] > upper:
                    outliers.append(i)
        return outliers

    def plot_intensity_profile(self):
        '''
        Plots the intensity profile of the current slice.
    
        This function is triggered whenever the current folder is updated or the slice number is changed
        The profile is calclulated using "calculate_profiles" and plotted on separate window

        '''
        if not hasattr(self,'profile_plot'):
            self.profile_plot = BgProfileWindow()
        self.profile_plot.fig.clear()
        profile_foreground, profile_background = self.calculate_profiles()
        self.profile_plot.updatefig(    
                foreground=profile_foreground[1:], 
                background=profile_background[1:], 
                mouse=self.mouse_name, 
                z_idx =self.cb.currentIndex(),
                sig=self.slider

                )
        self.profile_plot.draw()


    def showDialog(self,msg=''):
        '''
        This is a function that shows a dialog box.
        It takes a single argument, which is the message to be displayed.
        '''
        msgBox = QMessageBox()
        msgBox.setText(msg)
        msgBox.setWindowTitle("MMAR Warning!")
        msgBox.setStandardButtons(QMessageBox.Ok)
        msgBox.exec()

    def exitapp(self):
        self.mask_fig.close()
        self.profile_plot.close()
        QCoreApplication.instance().quit()


class BgProfileWindow(FigureCanvas):
    def __init__(self, width=6, height=4, dpi=100):

        self.fig = Figure(figsize=(width, height), dpi=dpi)
        FigureCanvas.__init__(self, self.fig)

        FigureCanvas.setSizePolicy(self,
                                   QSizePolicy.Expanding,
                                   QSizePolicy.Expanding)
        FigureCanvas.updateGeometry(self)
        
    def updatefig(self, foreground, background, mouse=None, z_idx=None, sig=None):
        self.sig = sig
        xx = np.arange(1, len(foreground)+1)
        self.fig.clear()
        self.axes = self.fig.add_subplot(111)
        self.axes.plot(xx, foreground, marker='o', color='dodgerblue', label='Foreground')
        self.axes.title.set_text(mouse+' Slice #'+str(z_idx))
        self.axes.title.set_fontsize(10)
        self.axes.set_xlabel('Slice Number')
        self.axes.set_ylabel('Foreground Intensity', color='dodgerblue')
        self.axes2 = self.axes.twinx()

        self.axes2.plot(xx, background, marker='o', color='darkorange', label='Background')
        self.axes2.set_ylabel('Background Intensity', color='darkorange')
        self.show()

        self.crs = mplcursors.cursor([self.axes, self.axes2],hover=True)
        self.crs.connect("add", lambda sel: sel.annotation.set_text(
            '{:.0f} : {:.2f}'.format(sel.target[0], sel.target[1])))
        self.fig.canvas.mpl_connect('button_press_event', self.onclick)

    def onclick(self, event):
        ix, iy = event.xdata, event.ydata
        self.sig.setValue(round(ix))

def run_gui():
    app = QApplication(sys.argv)
    path = os.path.join(os.path.dirname(sys.modules[__name__].__file__), 'icon.png')
    app.setWindowIcon(QIcon(path))
    ex = App()
    sys.exit(app.exec_())

if __name__ == '__main__':
    run_gui()



import ipywidgets as ipyw
from skimage.color import rgb2gray
from skimage.filters import threshold_triangle, threshold_otsu
from skimage.morphology import binary_closing
from scipy import ndimage
import numpy as np
import matplotlib.pyplot as plt

def norm_im(data):
    '''
    normalizes 2d array to [0, 1]
    '''
    return (data - np.min(data)) / (np.max(data) - np.min(data))

def estimate_background(image, threshold_algo='triangle_otsu',
                        threshold=None):
    """
    Estimates the background and pixel intesnity background theshold for an
    image. Note intensity is calculated using skimage.color.rgb2grey
    Parameters
    ----------
    image: array-like, (height, width, n_channels)
        Image whose background to estimate.
    threshold_algo: str, ['otsu', 'triangle', 'triangle_otsu']
        Thresholding algorithm to estimate the background.
        'otsu': skimage.filters.threshold_otsu
        'triangle': skimage.filters.threshold_triangle
        'triangle_otsu': .9 * triangle + .1 * otsu
    threshold: None, float, int
        User provided threshold. If None, will be estimated using one
        of the thresholding algorithms.
    Output
    ------
    background_mask, threshold
    background_mask: array-like, (height, width)
        The True/False mask of estimated background pixels.
    threshold: float
        The (lower bound) backgound threshold intensity for the image.
    """

    grayscale_image = rgb2gray(image)

    if threshold is None:
        if threshold_algo == 'otsu':
            threshold = threshold_otsu(grayscale_image)

        elif threshold_algo == 'triangle':
            threshold = threshold_triangle(grayscale_image)

        elif threshold_algo == 'triangle_otsu':
            triangle = threshold_triangle(grayscale_image)
            otsu = threshold_otsu(grayscale_image)

            threshold = .9 * triangle + .1 * otsu

        else:
            raise ValueError('threshold_algo = {} is invalid argument'.\
                  format(threshold_algo))

    background_mask = grayscale_image > threshold

    return close_and_fill(background_mask), threshold


def close_and_fill(image):
    '''
    this function closes and fills a binary image
    '''
    return ndimage.binary_fill_holes(binary_closing(image))


class ImageSliceViewer3D:
    """ 
    ImageSliceViewer3D is for viewing volumetric image slices in jupyter or
    ipython notebooks. 
    
    User can interactively change the slice plane selection for the image and 
    the slice plane being viewed. 

    Argumentss:
    Volume = 3D input image
    figsize = default(8,8), to set the size of the figure
    cmap = default('plasma'), string for the matplotlib colormap. You can find 
    more matplotlib colormaps on the following link:
    https://matplotlib.org/users/colormaps.html
    
    """
    
    def __init__(self, volume, figsize=(8,8), cmap='plasma'):
        self.volume = volume
        self.figsize = figsize
        self.cmap = cmap
        self.v = [np.min(volume), np.max(volume)]
        
        # Call to select slice plane
        ipyw.interact(self.view_selection, view=ipyw.RadioButtons(
            options=['x-y','y-z', 'z-x'], value='y-z', 
            description='Slice plane selection:', disabled=False,
            style={'description_width': 'initial'}))
    
    def view_selection(self, view):
        # Transpose the volume to orient according to the slice plane selection
        orient = {"y-z":[1,2,0], "z-x":[2,0,1], "x-y": [0,1,2]}
        self.vol = np.transpose(self.volume, orient[view])
        maxZ = self.vol.shape[2] - 1
        
        # Call to view a slice within the selected slice plane
        ipyw.interact(self.plot_slice, 
            z=ipyw.IntSlider(min=0, max=maxZ, step=1, continuous_update=False, 
            description='Image Slice:'))
        
    def plot_slice(self, z):
        # Plot slice for the given plane and slice
        self.fig = plt.figure(figsize=self.figsize)
        plt.imshow(self.vol[:,:,z], cmap=plt.get_cmap(self.cmap), 
            vmin=self.v[0], vmax=self.v[1])



import numpy as np
import matplotlib.pyplot as plt
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler
from sklearn.svm import SVC
from sklearn.linear_model import LogisticRegression
from sklearn.ensemble import RandomForestClassifier, AdaBoostClassifier, GradientBoostingClassifier
from sklearn.tree import DecisionTreeClassifier
from sklearn.naive_bayes import GaussianNB
from sklearn.metrics import balanced_accuracy_score, confusion_matrix, classification_report, roc_auc_score, roc_curve
from imblearn.over_sampling import SMOTE, ADASYN
from imblearn.under_sampling import RandomUnderSampler
from imblearn.pipeline import Pipeline

class MotionFrameDetector:
    def __init__(self, X, Y, train_balanced=False):
        self.X = X
        self.Y = Y
        
        self.train_balanced = train_balanced

        self.X_train, self.X_test, self.Y_train, self.Y_test = train_test_split(self.X, self.Y, test_size=0.2, random_state=23)
        
        sc = StandardScaler()
        self.X_train = sc.fit_transform(self.X_train)
        self.X_test = sc.transform(self.X_test)

        self.classifiers = [
            LogisticRegression(), 
            SVC(kernel='linear',probability=True), 
            SVC(probability=True), 
            RandomForestClassifier(),
            AdaBoostClassifier(DecisionTreeClassifier(max_depth=1),
                         algorithm="SAMME",
                         n_estimators=200),
            GradientBoostingClassifier(n_estimators=100, learning_rate=.1, max_depth=2),
            GaussianNB()
            ]
        
        self.classifier_names = [
            'Logistic Regression', 
            'SVM with linear kernel', 
            'SVM', 
            'Random Forest', 
            'AdaBoosted decision tree', 
            'Gradient Boosted Decision Tree',
            'Gaussian Naive Bayes',
            ]
        
        self.classifier_accuracies = []
        self.classifier_predictions = []
        self.classifier_probabilities = []
        self.classifier_confusion_matrices = []
        self.classifier_classification_reports = []
        self.classifier_roc_auc_scores = []
        self.classifier_roc_curves = []
        self.classifier_balanced_accuracy = []

    def train(self):
        if self.train_balanced:
            X, y = self.balance_data(self.X_train, self.Y_train)
        else:
            X, y = self.X_train, self.Y_train
        for classifier, classifier_name in zip(self.classifiers, self.classifier_names):
            classifier.fit(X, y)
            self.classifier_accuracies.append(classifier.score(self.X_test, self.Y_test))
            self.classifier_predictions.append(classifier.predict(self.X_test))
            self.classifier_probabilities.append(classifier.predict_proba(self.X_test))
            self.classifier_confusion_matrices.append(confusion_matrix(self.Y_test, self.classifier_predictions[-1]))
            self.classifier_classification_reports.append(classification_report(self.Y_test, self.classifier_predictions[-1]))
            self.classifier_roc_auc_scores.append(roc_auc_score(self.Y_test, self.classifier_probabilities[-1][:, 1]))
            self.classifier_roc_curves.append(roc_curve(self.Y_test, self.classifier_probabilities[-1][:, 1]))
            self.classifier_balanced_accuracy.append(balanced_accuracy_score(self.Y_test, self.classifier_predictions[-1]))
    
    def balance_data(self, X, y):
        '''
        balance the minority class 
        ref: https://machinelearningmastery.com/smote-oversampling-for-imbalanced-classification/
        '''
        # over = SMOTE(sampling_strategy=0.5)
        over = ADASYN(sampling_strategy=0.5)
        under = RandomUnderSampler(sampling_strategy=0.5)
        steps = [('o', over), ('u', under)]
        pipeline = Pipeline(steps=steps)
        X, y =pipeline.fit_resample(X, y)
        return X, y

    def print_results(self):
        for classifier_name, classifier_accuracy, classifier_balanced_accuracy, classifier_roc_auc_score in zip(self.classifier_names, self.classifier_accuracies, self.classifier_balanced_accuracy, self.classifier_roc_auc_scores):
            print('Accuracy of {}: {:.2f}%'.format(classifier_name, classifier_accuracy*100))
            print('Balanced Accuracy of {}: {:.2f}%'.format(classifier_name, classifier_balanced_accuracy*100))
            print('ROC AUC score of {}: {:.2f}%'.format(classifier_name, classifier_roc_auc_score*100))
            print('\n')

    def plot_confusion_matrices(self):
        fig = plt.figure(dpi=75, figsize=(12,12))

        k = 1
        for classifier_name, classifier_confusion_matrix in zip(self.classifier_names, self.classifier_confusion_matrices):
            plt.subplot(4,2,k)
            plt.imshow(classifier_confusion_matrix, interpolation='nearest', cmap=plt.cm.Blues)
            plt.title('Confusion matrix for {}'.format(classifier_name))
            # plt.colorbar()
            tick_marks = np.arange(2)
            plt.xticks(tick_marks, ['0', '1'], rotation=45)
            plt.yticks(tick_marks, ['0', '1'])
            # plt.tight_layout()
            plt.ylabel('True label')
            plt.xlabel('Predicted label')
            k+=1
        plt.show()

    def plot_roc_curves(self):
        plt.figure()
        for classifier_name, classifier_roc_curve in zip(self.classifier_names, self.classifier_roc_curves):
            plt.plot(classifier_roc_curve[0], classifier_roc_curve[1], label=classifier_name)
        plt.plot([0, 1], [0, 1], 'k--', label='')
        plt.axis([0, 1, 0, 1])
        plt.xlabel('False Positive Rate')
        plt.ylabel('True Positive Rate')
        plt.title('ROC curves')
        plt.legend()
        plt.show()

    def plot_roc_auc_scores(self):
        plt.figure()
        plt.bar(range(len(self.classifier_roc_auc_scores)), self.classifier_roc_auc_scores, align='center')
        plt.xticks(range(len(self.classifier_roc_auc_scores)), self.classifier_names, rotation=45, ha='right')
        plt.title('ROC AUC score')
        plt.ylim((0.75, 1))
        plt.show()

    def plot_balanced_scores(self):
        plt.figure()
        plt.bar(range(len(self.classifier_balanced_accuracy)), self.classifier_balanced_accuracy, align='center')
        plt.xticks(range(len(self.classifier_balanced_accuracy)), self.classifier_names, rotation=45, ha='right')
        plt.title('Balanced Accuracy')
        plt.ylim((0.5, 1))
        plt.show()


#%%
import torch
import numpy as np
from dataset import load_data
from metrics import dice_score
from torchvision import transforms
from PIL import Image
from utils import cvt1to3channels, normalize_image
import matplotlib.pyplot as plt


def get_dice_val(model, src_val, msk_val, fnames_val, trans):
    max_slice_num = max([i[1] for i in fnames_val])
    dice_val = {}

    for i_slice in range(max_slice_num):
        print(f'Starting slice number: {i_slice}')
        dice_val[i_slice] = []
        idx_slice = [i for i, f in enumerate(fnames_val) if f[1]==i_slice]
        inputs_all = torch.empty([len(idx_slice),3,256,256])
        labels_all = torch.empty([len(idx_slice),1,256,256])
        for ii, idx1 in enumerate(idx_slice):
            inputs, labels = src_val[idx1], msk_val[idx1]
            inputs = trans(Image.fromarray(cvt1to3channels(inputs)))
            labels = trans(Image.fromarray(np.uint8(normalize_image(labels))))
            inputs_all[ii,:,:,:] = inputs
            labels_all[ii,:,:,:] = labels
        
        model.eval()
        with torch.no_grad():
            predict = model(inputs_all)

        # Saving metrics
        dice_val[i_slice] = dice_score(labels_all, predict).item()
        print(f'Validation dice score for slice number: {i_slice} of {modality} was {dice_val[i_slice]}')
    return dice_val

dice_modality = {}
modalities = ['dti', 'fmri', 'anatomical', 'noddi']
for modality in modalities:
    skpath='/home/ghanba/ssNET/SkullStripperUNET/weights/'+modality+'.pth'
    # skpath='/home/ghanba/ssNET/skull-stripper/paper_weights/skull-stripper-paper.pth'
    data_path = '/projects/compsci/USERS/frohoz/msUNET/train/dataset/'
    validation_portion=.2
    _, _, src_val, msk_val, fnames_train, fnames_val = load_data(data_path,
    num_train_sample=80, modality=modality)
    Nt = len(_) # number of trian samples

    model = torch.hub.load('mateuszbuda/brain-segmentation-pytorch', 'unet',
                        in_channels=3, out_channels=1, init_features=32,
                        pretrained=False)
    model.load_state_dict(torch.load(skpath, map_location=torch.device('cpu')))

    trans = transforms.Compose([transforms.Resize((225,225)),transforms.CenterCrop(256), transforms.ToTensor()])
    print(f'Calculating dice score for slices in {modality}')
    dice_modality[modality] = get_dice_val(model, src_val, msk_val, fnames_val, trans)

#%%
for modality in modalities:
    plt.plot(dice_modality[modality].values(),'.')

plt.xlabel('Slice Number')
plt.ylabel('Mean Dice Coeff')
plt.legend(modalities)
plt.ylim((.85, 1))
plt.savefig('results_.pdf')
plt.show()
# %%
i_slice = 7
idx_slice = [i for i, f in enumerate(fnames_val) if f[1]==i_slice]

for idx1 in idx_slice[:10]:
    inputs, labels = src_val[idx1], msk_val[idx1]
    labels = trans(Image.fromarray(np.uint8(normalize_image(labels))))
    # predict = model(trans(Image.fromarray(cvt1to3channels(inputs))).unsqueeze(0))
    model.eval()
    with torch.no_grad():
        predict = model(trans(Image.fromarray(cvt1to3channels(inputs))).unsqueeze(0))
    print(f'dice score: {dice_score(predict[0,0,:,:], labels[0,:,:])}')

    plt.subplot(1,2,1)
    plt.title(f'max of labels: {np.max(labels[0,:,:].numpy())}')
    plt.imshow(labels[0,:,:])

    plt.subplot(1,2,2)
    plt.title(f'max of predict: {np.max(predict[0,0,:,:].detach().numpy())}')
    plt.imshow(predict.detach().numpy()[0,0,:,:])
    plt.show()

# %%



#%%
import os
import sys
import numpy as np
import pandas as pd
import tifffile as tiff
import matplotlib.pyplot as plt
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler
from sklearn.svm import SVC
from sklearn.linear_model import LogisticRegression
from sklearn.ensemble import RandomForestClassifier, AdaBoostClassifier, GradientBoostingClassifier
from sklearn.tree import DecisionTreeClassifier
from sklearn.naive_bayes import GaussianNB
from sklearn.metrics import balanced_accuracy_score, confusion_matrix, classification_report, roc_auc_score, roc_curve
from skimage.restoration import estimate_sigma
from imblearn.over_sampling import SMOTE
from imblearn.under_sampling import RandomUnderSampler
from imblearn.pipeline import Pipeline
sys.path.append(os.path.join(os.path.dirname(__file__), '..', 'mmar'))
import process_slice

class Classifier:
    def __init__(self, X, Y, train_balanced=False):
        self.X = X
        self.Y = Y
        
        self.train_balanced = train_balanced

        self.X_train, self.X_test, self.Y_train, self.Y_test = train_test_split(self.X, self.Y, test_size=0.2, random_state=23)
        
        sc = StandardScaler()
        self.X_train = sc.fit_transform(self.X_train)
        self.X_test = sc.transform(self.X_test)

        self.classifiers = [
            LogisticRegression(), 
            SVC(kernel='linear',probability=True), 
            SVC(probability=True), 
            RandomForestClassifier(),
            AdaBoostClassifier(DecisionTreeClassifier(max_depth=1),
                         algorithm="SAMME",
                         n_estimators=200),
            GradientBoostingClassifier(n_estimators=100, learning_rate=.1, max_depth=2),
            GaussianNB()
            ]
        
        self.classifier_names = [
            'Logistic Regression', 
            'SVM with linear kernel', 
            'SVM', 
            'Random Forest', 
            'AdaBoosted decision tree', 
            'Gradient Boosted Decision Tree',
            'Gaussian Naive Bayes',
            ]
        
        self.classifier_accuracies = []
        self.classifier_predictions = []
        self.classifier_probabilities = []
        self.classifier_confusion_matrices = []
        self.classifier_classification_reports = []
        self.classifier_roc_auc_scores = []
        self.classifier_roc_curves = []
        self.classifier_balanced_accuracy = []

    def train(self):
        if self.train_balanced:
            X, y = self.balance_data(self.X_train, self.Y_train)
        else:
            X, y = self.X_train, self.Y_train
        for classifier, classifier_name in zip(self.classifiers, self.classifier_names):
            classifier.fit(X, y)
            self.classifier_accuracies.append(classifier.score(self.X_test, self.Y_test))
            self.classifier_predictions.append(classifier.predict(self.X_test))
            self.classifier_probabilities.append(classifier.predict_proba(self.X_test))
            self.classifier_confusion_matrices.append(confusion_matrix(self.Y_test, self.classifier_predictions[-1]))
            self.classifier_classification_reports.append(classification_report(self.Y_test, self.classifier_predictions[-1]))
            self.classifier_roc_auc_scores.append(roc_auc_score(self.Y_test, self.classifier_probabilities[-1][:, 1]))
            self.classifier_roc_curves.append(roc_curve(self.Y_test, self.classifier_probabilities[-1][:, 1]))
            self.classifier_balanced_accuracy.append(balanced_accuracy_score(self.Y_test, self.classifier_predictions[-1]))
    
    def balance_data(self, X, y):
        '''
        balance the minority class 
        ref: https://machinelearningmastery.com/smote-oversampling-for-imbalanced-classification/
        '''
        over = SMOTE(sampling_strategy=0.5)
        under = RandomUnderSampler(sampling_strategy=0.5)
        steps = [('o', over), ('u', under)]
        pipeline = Pipeline(steps=steps)
        X, y =pipeline.fit_resample(X, y)
        return X, y

    def print_results(self):
        for classifier_name, classifier_accuracy, classifier_balanced_accuracy, classifier_roc_auc_score in zip(self.classifier_names, self.classifier_accuracies, self.classifier_balanced_accuracy, self.classifier_roc_auc_scores):
            print('Accuracy of {}: {:.2f}%'.format(classifier_name, classifier_accuracy*100))
            print('Balanced Accuracy of {}: {:.2f}%'.format(classifier_name, classifier_balanced_accuracy*100))
            print('ROC AUC score of {}: {:.2f}%'.format(classifier_name, classifier_roc_auc_score*100))
            print('\n')

    def plot_confusion_matrices(self):
        fig = plt.figure(dpi=150, figsize=(20,20))

        k = 1
        for classifier_name, classifier_confusion_matrix in zip(self.classifier_names, self.classifier_confusion_matrices):
            plt.subplot(2,5,k)
            plt.imshow(classifier_confusion_matrix, interpolation='nearest', cmap=plt.cm.Blues)
            plt.title('Confusion matrix for {}'.format(classifier_name))
            # plt.colorbar()
            tick_marks = np.arange(2)
            plt.xticks(tick_marks, ['0', '1'], rotation=45)
            plt.yticks(tick_marks, ['0', '1'])
            # plt.tight_layout()
            plt.ylabel('True label')
            plt.xlabel('Predicted label')
            k+=1
        plt.show()

    def plot_roc_curves(self):
        plt.figure()
        for classifier_name, classifier_roc_curve in zip(self.classifier_names, self.classifier_roc_curves):
            plt.plot(classifier_roc_curve[0], classifier_roc_curve[1], label=classifier_name)
        plt.plot([0, 1], [0, 1], 'k--', label='')
        plt.axis([0, 1, 0, 1])
        plt.xlabel('False Positive Rate')
        plt.ylabel('True Positive Rate')
        plt.title('ROC curves')
        plt.legend()
        plt.show()

    def plot_roc_auc_scores(self):
        plt.figure()
        plt.bar(range(len(self.classifier_roc_auc_scores)), self.classifier_roc_auc_scores, align='center')
        plt.xticks(range(len(self.classifier_roc_auc_scores)), self.classifier_names, rotation=45, ha='right')
        plt.title('ROC AUC score')
        plt.ylim((0.75, 1))
        plt.show()

    def plot_balanced_scores(self):
        plt.figure()
        plt.bar(range(len(self.classifier_balanced_accuracy)), self.classifier_balanced_accuracy, align='center')
        plt.xticks(range(len(self.classifier_balanced_accuracy)), self.classifier_names, rotation=45, ha='right')
        plt.title('Balanced Accuracy')
        plt.ylim((0.5, 1))
        plt.show()

def feature_extractor(image, mask):
    # individual image feature extractor
    features = [
        np.mean(image), 
        np.std(image), 

        np.mean(image[mask]), 
        np.std(image[mask]), 
        np.mean(image[mask]) / np.std(image[mask]),
        
        np.mean(image[mask==False]), 
        np.std(image[mask==False]), 
        np.mean(image[mask==False]) / np.std(image[mask==False]),
        
        estimate_sigma(image),
        
        psnr(image, image*mask),
        psnr(image, image*(1-mask))
        ]
    return features

def extract_features_for_all(parent_folder, mouse_list):

    features = {}
    for mouse in mouse_list:
        print(mouse)
        tif_file = os.path.join(parent_folder, mouse, 'dti_large_0.tif')
        if not os.path.isfile(tif_file):
            features[mouse]=[]
            continue

        img = tiff.imread(tif_file)[1:,:,:,:]
        nframes, nslices, nx, ny = img.shape
        nf_static = 11 # number of static features
        nf_dynamic = 7 # number of dynamic features -- has some information beyond 2D image
        features[mouse] = np.empty((nframes, nslices, nf_static+nf_dynamic))
        for i in range(nframes):
            for j in range(nslices):
                mask = process_slice.create_mask(img[i,j,:,:], fg=False)[0]
                features[mouse][i,j,:nf_static] = feature_extractor(img[i,j,:,:], mask)
            # add general slice feature to each individual frame // based on Median Absloute Deviation

        for j in range(nslices):
            allframes_data_median = np.median(features[mouse][:,j,0])
            allframes_fg_data_median = np.median(features[mouse][:,j,2])
            allframes_bg_data_median = np.median(features[mouse][:,j,4])
            allframes_MAD = np.median([np.abs(features[mouse][:,j,0] - allframes_data_median)])
            allframes_fg_MAD = np.median([np.abs(features[mouse][:,j,2] - allframes_fg_data_median)])
            allframes_bg_MAD = np.median([np.abs(features[mouse][:,j,4] - allframes_bg_data_median)])
            for i in range(nframes):
                features[mouse][i,j,nf_static] = allframes_fg_data_median
                features[mouse][i,j,nf_static+1] = allframes_fg_MAD
                features[mouse][i,j,nf_static+2] = allframes_bg_data_median
                features[mouse][i,j,nf_static+3] = allframes_bg_MAD
                features[mouse][i,j,nf_static+4] = allframes_MAD
                features[mouse][i,j,nf_static+5] = allframes_data_median
                features[mouse][i,j,nf_static+6] = j    

            
    return features

def psnr(img1, img2):
    '''
    Peak Signal to Noise Ratio
    '''
    mse = np.mean((img1 - img2) ** 2)
    return 20*np.log10(np.max(img1) / np.sqrt(mse))

def find_excel_file(folder_path):
    file_list = os.listdir(folder_path)
    excel_file_list = []
    for file in file_list:
        if file.endswith(".xlsx") and not file.startswith("~"):
            excel_file_list.append(file)
    return excel_file_list

def get_rejected_frames(parent_folder, mouse_list):

    rejected = {}
    for mouse in mouse_list:
        rejected_frames_file = find_excel_file(os.path.join(parent_folder,mouse))
        assert len(rejected_frames_file)==1, f"more than one excel file in {mouse}"
        df = pd.read_excel(os.path.join(parent_folder,mouse,rejected_frames_file[0]))
        rejected[mouse] = [[]]*17
        for i, val in enumerate(df[df.columns[0]]):
            if isinstance(val, int): 
                if isinstance(df.iloc[i,1], str):
                    rejected[mouse][val-1]=[int(i)-2 for i in df.iloc[i,1].split(',') if len(i)>0]
                elif isinstance(df.iloc[i,1], int):
                    rejected[mouse][val-1]=df.iloc[i,1]-2
    return rejected


def set_to_one(A, B):
    for i in range(len(A)):
        if A[i] != []:
            if isinstance(A[i], int):
                B[A[i]][i] = 1
                continue
            for j in A[i]:
                B[j][i] = 1
    return B

parent_folder = '/Users/ghanba/Documents/ssif_data_jax/'
mouse_list = [f for f in os.listdir(parent_folder) if not f.startswith('.')]
mouse_list.sort()

mouse_list = mouse_list[:10]

rejected = get_rejected_frames(parent_folder, mouse_list)

features = extract_features_for_all(parent_folder, mouse_list)

X_all = []
Y_all = []
for mouse in mouse_list:
    X = features[mouse]
    if isinstance(X_all, list):
        X_all = np.expand_dims(X, axis=0)
    else:
        X_all = np.concatenate((X_all, np.expand_dims(X, axis=0)), axis=0)
    
    Y = set_to_one(rejected[mouse], np.zeros((X.shape[0], X.shape[1])))
    if isinstance(Y_all, list):
        Y_all = np.expand_dims(Y, axis=0)
    else:
        Y_all = np.concatenate((Y_all, np.expand_dims(Y, axis=0)), axis=0)
    
X_all = X_all[:,:,:,:]
X = X_all.reshape(-1, X_all.shape[-1])
Y = Y_all.reshape(-1)


#%% scatter plot of the dataset
from collections import Counter
def plot_features(X, Y, n):
    fig = plt.figure(dpi=150, figsize=(20,20))

    k = 1

    for j in range(X.shape[1]):
        plt.subplot(5,5,k)
        counter = Counter(Y)
        for label, _ in counter.items():
            row_ix = np.where(Y == label)[0]
            plt.scatter(X[row_ix, n], X[row_ix, j], label=str(label))
        k+=1

    plt.legend()
    plt.show()
plot_features(X, Y, 11)
#%%

#%%
model = Classifier(X, Y, train_balanced=True)
model.train()
model.print_results()
model.plot_confusion_matrices()
model.plot_roc_curves()
model.plot_roc_auc_scores()
model.plot_balanced_scores()




import numpy as np


def cvt1to3channels(one_channel):
    return np.stack((one_channel,)*3, axis=-1)

def normalize_image(image):
    return 255*((image - np.min(image)) / (np.max(image) - np.min(image)))




from setuptools import setup, find_packages

with open('requirements.txt') as f:
    required = f.read().splitlines()

setup(
    name='MotionArtifactRemoval',
    version='1.0.11',    
    description='Motion Artifact Removal',
    author='Jim Peterson, Abed Ghanbari',
    url='https://www.jax.org',
    author_email='abed.ghanbari@jax.org',
    license='BSD 2-clause',
    packages=find_packages(),
    install_requires=required,

    classifiers=[
        'Development Status :: 1 - Planning',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',  
        'Operating System :: POSIX :: Linux',        
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    include_package_data=True,
    package_data={'ssUNET': ['weights/*.pth']},
)



# Copyright The Jackson Laboratory, 2021
# authors: Jim Peterson, Abed Ghanbari
import os

class message_log:
    cumulative_message = ""
    verbose = True

    def __init__(self, verbose):
        self.verbose = verbose
        return

    def add_msg(self, msg, is_error_msg=False):
        self.cumulative_message += msg
        self.cumulative_message += os.linesep
        if self.verbose or is_error_msg:
            print(msg)
        return


    def save_log(self, file_path):
        with open(file_path, 'w') as f:
            f.write(self.cumulative_message)
        return



# Copyright The Jackson Laboratory, 2021
# authors: Jim Peterson, Abed Ghanbari

import numpy as np
from scipy import ndimage
from skimage.filters import threshold_triangle, threshold_otsu
from skimage.morphology import binary_closing
from skimage.measure import label

def create_mask(image, threshold_algo='triangle_otsu', threshold=None, fg=True):
    """
    Estimates the background and pixel intesnity background theshold for an
    image. Note intensity is calculated using skimage.color.rgb2grey
    Parameters
    ----------
    image: array-like, (height, width)
        Image from which the mask is created.
    threshold_algo: str, ['otsu', 'triangle', 'triangle_otsu']
        Thresholding algorithm to estimate the background.
        'otsu': skimage.filters.threshold_otsu
        'triangle': skimage.filters.threshold_triangle
        'triangle_otsu': .9 * triangle + .1 * otsu
    threshold: None, float, int
        User provided threshold. If None, will be estimated using one
        of the thresholding algorithms.
    Output
    ------
    background_mask, thresholded first frame
    background_mask: array-like, (height, width)
        The True/False mask of estimated background pixels.
    threshold: float
        The lower-bound backgound threshold, or upper-bound foreground intensity for the image.
    """

    if threshold is None:
        if threshold_algo == 'otsu':
            threshold = threshold_otsu(image)

        elif threshold_algo == 'triangle':
            threshold = threshold_triangle(image)

        elif threshold_algo == 'triangle_otsu':
            triangle = threshold_triangle(image)
            otsu = threshold_otsu(image)

            threshold = .9 * triangle + .1 * otsu

        else:
            raise ValueError('threshold_algo = {} is invalid argument'.format(threshold_algo))
    if fg:
        mask = image < threshold
    else:
        mask = image > threshold

    refined_mask = get_largest_island(close_and_fill(mask))
    
    return refined_mask, threshold



def close_and_fill(image):
    '''
    This function closes and fills a binary image.
    
    Parameters
    ----------
    image : numpy.ndarray
        A binary image.
        
    Returns
    -------
    numpy.ndarray
        A binary image.
    '''
    return ndimage.binary_fill_holes(binary_closing(image))

def get_largest_island(binary_image):
    labeled_image, num_features = label(binary_image,return_num=True)
    max_area = 0
    max_label = 0
    for i in range(1, num_features + 1):
        area = np.sum(labeled_image == i)
        if area > max_area:
            max_area = area
            max_label = i
    return (labeled_image == max_label)


import torch.nn as nn


class DiceLoss(nn.Module):

    def __init__(self):
        super(DiceLoss, self).__init__()
        self.smooth = 1.0

    def forward(self, y_pred, y_true):
        assert y_pred.size() == y_true.size()
        y_pred = y_pred[:, 0].contiguous().view(-1)
        y_true = y_true[:, 0].contiguous().view(-1)
        intersection = (y_pred * y_true).sum()
        dsc = (2. * intersection + self.smooth) / (
            y_pred.sum() + y_true.sum() + self.smooth
        )
        return 1. - dsc
